@@ -110,7 +110,7 @@ int copy_timestamp(const char *src_filename, const char *dst_filename)
 //////////////////////////////////////////////////////////////////////////////
 // This function parses a filename (with or without full path) and returns  //
 // a pointer to the extension (including the "."). If no extension is found //
-// then NULL is returned. Extensions with more than 3 letters don't count.  //
+// then NULL is returned. Extensions with more than 4 letters don't count.  //
 //////////////////////////////////////////////////////////////////////////////
 
 #if defined(_WIN32)
@@ -125,7 +125,7 @@ char *filespec_ext (char *filespec)
             return NULL;
 
         if (*cp == '.') {
-            if (strlen (cp) > 1 && strlen (cp) <= 4)
+            if (strlen (cp+1) && strlen (cp+1) <= 4)
                 return cp;
             else
                 return NULL;
@@ -147,7 +147,7 @@ char *filespec_ext (char *filespec)
             return NULL;
 
         if (*cp == '.') {
-            if (strlen (cp) > 1 && strlen (cp) <= 4)
+            if (strlen (cp+1) && strlen (cp+1) <= 4)
                 return cp;
             else
                 return NULL;
@@ -670,6 +670,21 @@ int64_t DoGetFileSize (FILE *hFile)
     return (int64_t)Size.QuadPart;
 }
 
+int64_t DoGetFilePosition (FILE *hFile)
+{
+    return _ftelli64 (hFile);
+}
+
+int DoSetFilePositionAbsolute (FILE *hFile, int64_t pos)
+{
+    return _fseeki64 (hFile, pos, SEEK_SET);
+}
+
+int DoSetFilePositionRelative (FILE *hFile, int64_t pos, int mode)
+{
+    return _fseeki64 (hFile, pos, mode);
+}
+
 #else
 
 int64_t DoGetFileSize (FILE *hFile)
@@ -682,23 +697,23 @@ int64_t DoGetFileSize (FILE *hFile)
     return (int64_t) statbuf.st_size;
 }
 
-#endif
-
-uint32_t DoGetFilePosition (FILE *hFile)
+int64_t DoGetFilePosition (FILE *hFile)
 {
     return ftell (hFile);
 }
 
-int DoSetFilePositionAbsolute (FILE *hFile, uint32_t pos)
+int DoSetFilePositionAbsolute (FILE *hFile, int64_t pos)
 {
     return fseek (hFile, pos, SEEK_SET);
 }
 
-int DoSetFilePositionRelative (FILE *hFile, int32_t pos, int mode)
+int DoSetFilePositionRelative (FILE *hFile, int64_t pos, int mode)
 {
     return fseek (hFile, pos, mode);
 }
 
+#endif
+
 // if ungetc() is not available, a seek of -1 is fine also because we do not
 // change the byte read.
 
