@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////////////////////
 //                           **** WAVPACK ****                            //
 //                  Hybrid Lossless Wavefile Compressor                   //
-//              Copyright (c) 1998 - 2013 Conifer Software.               //
+//                Copyright (c) 1998 - 2016 David Bryant.                 //
 //                          All Rights Reserved.                          //
 //      Distributed under the BSD Software License (see license.txt)      //
 ////////////////////////////////////////////////////////////////////////////
@@ -29,9 +29,9 @@
 // function which handles the correction file transparently, in this case it
 // is the responsibility of the caller to be aware of correction files.
 
-static uint32_t seek_final_index (WavpackStreamReader *reader, void *id);
+static int seek_eof_information (WavpackContext *wpc, int64_t *final_index, int get_wrapper);
 
-WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id, void *wvc_id, char *error, int flags, int norm_offset)
+WavpackContext *WavpackOpenFileInputEx64 (WavpackStreamReader64 *reader, void *wv_id, void *wvc_id, char *error, int flags, int norm_offset)
 {
     WavpackContext *wpc = malloc (sizeof (WavpackContext));
     WavpackStream *wps;
@@ -48,7 +48,7 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
     wpc->wv_in = wv_id;
     wpc->wvc_in = wvc_id;
     wpc->reader = reader;
-    wpc->total_samples = (uint32_t) -1;
+    wpc->total_samples = -1;
     wpc->norm_offset = norm_offset;
     wpc->max_streams = OLD_MAX_STREAMS;     // use this until overwritten with actual number
     wpc->open_flags = flags;
@@ -67,7 +67,6 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
     }
 #endif
 
-#ifndef VER4_ONLY
     if (wpc->reader->read_bytes (wpc->wv_in, &first_byte, 1) != 1) {
         if (error) strcpy (error, "can't read all of WavPack file!");
         return WavpackCloseFile (wpc);
@@ -75,9 +74,14 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
 
     wpc->reader->push_back_byte (wpc->wv_in, first_byte);
 
-    if (first_byte == 'R')
+    if (first_byte == 'R') {
+#ifdef ENABLE_LEGACY
         return open_file3 (wpc, error);
+#else
+        if (error) strcpy (error, "this legacy WavPack file is deprecated, use version 4.80.0 to transcode");
+        return WavpackCloseFile (wpc);
 #endif
+    }
 
     wpc->streams = malloc ((wpc->num_streams = 1) * sizeof (wpc->streams [0]));
     if (!wpc->streams) {
@@ -116,30 +120,49 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
             return WavpackCloseFile (wpc);
         }
 
-        wps->init_done = FALSE;
-
-        if (wps->wphdr.block_samples && !(flags & OPEN_STREAMING)) {
-            if (wps->wphdr.block_index || wps->wphdr.total_samples == (uint32_t) -1) {
-                wpc->initial_index = wps->wphdr.block_index;
-                wps->wphdr.block_index = 0;
-
-                if (wpc->reader->can_seek (wpc->wv_in)) {
-                    uint32_t pos_save = wpc->reader->get_pos (wpc->wv_in);
-                    uint32_t final_index = seek_final_index (wpc->reader, wpc->wv_in);
-
-                    if (final_index != (uint32_t) -1)
-                        wpc->total_samples = final_index - wpc->initial_index;
-
-                    wpc->reader->set_pos_abs (wpc->wv_in, pos_save);
-                }
-            }
-            else
-                wpc->total_samples = wps->wphdr.total_samples;
+        // if block does not verify, flag error, free buffer, and continue
+        if (!WavpackVerifySingleBlock (wps->blockbuff, !(flags & OPEN_NO_CHECKSUM))) {
+            wps->wphdr.block_samples = 0;
+            free (wps->blockbuff);
+            wps->blockbuff = NULL;
+            wpc->crc_errors++;
+            continue;
         }
 
+        wps->init_done = FALSE;
+
+        if (wps->wphdr.block_samples) {
+            if (flags & OPEN_STREAMING)
+                SET_BLOCK_INDEX (wps->wphdr, 0);
+            else if (wpc->total_samples == -1) {
+                if (GET_BLOCK_INDEX (wps->wphdr) || GET_TOTAL_SAMPLES (wps->wphdr) == -1) {
+                    wpc->initial_index = GET_BLOCK_INDEX (wps->wphdr);
+                    SET_BLOCK_INDEX (wps->wphdr, 0);
+
+                    if (wpc->reader->can_seek (wpc->wv_in)) {
+                        int64_t final_index = -1;
+
+                        seek_eof_information (wpc, &final_index, FALSE);
+
+                        if (final_index != -1)
+                            wpc->total_samples = final_index - wpc->initial_index;
+                    }
+                }
+                else
+                    wpc->total_samples = GET_TOTAL_SAMPLES (wps->wphdr);
+            }
+        }
+        else if (wpc->total_samples == -1 && !GET_BLOCK_INDEX (wps->wphdr) && GET_TOTAL_SAMPLES (wps->wphdr))
+            wpc->total_samples = GET_TOTAL_SAMPLES (wps->wphdr);
+
         if (wpc->wvc_in && wps->wphdr.block_samples && (wps->wphdr.flags & HYBRID_FLAG)) {
-            wpc->file2len = wpc->reader->get_length (wpc->wvc_in);
-            wpc->wvc_flag = TRUE;
+            unsigned char ch;
+
+            if (wpc->reader->read_bytes (wpc->wvc_in, &ch, 1) == 1) {
+                wpc->reader->push_back_byte (wpc->wvc_in, ch);
+                wpc->file2len = wpc->reader->get_length (wpc->wvc_in);
+                wpc->wvc_flag = TRUE;
+            }
         }
 
         if (wpc->wvc_flag && !read_wvc_block (wpc)) {
@@ -159,18 +182,6 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
 
     wpc->config.flags &= ~0xff;
     wpc->config.flags |= wps->wphdr.flags & 0xff;
-    wpc->config.bytes_per_sample = (wps->wphdr.flags & BYTES_STORED) + 1;
-    wpc->config.float_norm_exp = wps->float_norm_exp;
-
-    wpc->config.bits_per_sample = (wpc->config.bytes_per_sample * 8) -
-        ((wps->wphdr.flags & SHIFT_MASK) >> SHIFT_LSB);
-
-    if (!wpc->config.sample_rate) {
-        if (!wps->wphdr.block_samples || (wps->wphdr.flags & SRATE_MASK) == SRATE_MASK)
-            wpc->config.sample_rate = 44100;
-        else
-            wpc->config.sample_rate = sample_rates [(wps->wphdr.flags & SRATE_MASK) >> SRATE_LSB];
-    }
 
     if (!wpc->config.num_channels) {
         wpc->config.num_channels = (wps->wphdr.flags & MONO_FLAG) ? 1 : 2;
@@ -180,66 +191,83 @@ WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id
     if ((flags & OPEN_2CH_MAX) && !(wps->wphdr.flags & FINAL_BLOCK))
         wpc->reduced_channels = (wps->wphdr.flags & MONO_FLAG) ? 1 : 2;
 
+    if (wps->wphdr.flags & DSD_FLAG) {
+#ifdef ENABLE_DSD
+        if (flags & OPEN_DSD_NATIVE) {
+            wpc->config.bytes_per_sample = 1;
+            wpc->config.bits_per_sample = 8;
+        }
+        else if (flags & OPEN_DSD_AS_PCM) {
+            wpc->decimation_context = decimate_dsd_init (wpc->reduced_channels ?
+                wpc->reduced_channels : wpc->config.num_channels);
+
+            wpc->config.bytes_per_sample = 3;
+            wpc->config.bits_per_sample = 24;
+        }
+        else {
+            if (error) strcpy (error, "not configured to handle DSD WavPack files!");
+            return WavpackCloseFile (wpc);
+        }
+#else
+        if (error) strcpy (error, "not configured to handle DSD WavPack files!");
+        return WavpackCloseFile (wpc);
+#endif
+    }
+    else {
+        wpc->config.bytes_per_sample = (wps->wphdr.flags & BYTES_STORED) + 1;
+        wpc->config.float_norm_exp = wps->float_norm_exp;
+
+        wpc->config.bits_per_sample = (wpc->config.bytes_per_sample * 8) -
+            ((wps->wphdr.flags & SHIFT_MASK) >> SHIFT_LSB);
+    }
+
+    if (!wpc->config.sample_rate) {
+        if (!wps->wphdr.block_samples || (wps->wphdr.flags & SRATE_MASK) == SRATE_MASK)
+            wpc->config.sample_rate = 44100;
+        else
+            wpc->config.sample_rate = sample_rates [(wps->wphdr.flags & SRATE_MASK) >> SRATE_LSB];
+    }
+
     return wpc;
 }
 
 // This function returns the major version number of the WavPack program
-// (or library) that created the open file. Currently, this can be 1 to 4.
+// (or library) that created the open file. Currently, this can be 1 to 5.
 // Minor versions are not recorded in WavPack files.
 
 int WavpackGetVersion (WavpackContext *wpc)
 {
     if (wpc) {
-#ifndef VER4_ONLY
+#ifdef ENABLE_LEGACY
         if (wpc->stream3)
             return get_version3 (wpc);
 #endif
-        return 4;
+        return wpc->version_five ? 5 : 4;
     }
 
     return 0;
 }
 
-// This function is used to seek to end of a file to determine its actual
-// length in samples by reading the last header block containing data.
-// Currently, all WavPack files contain the sample length in the first block
-// containing samples, however this might not always be the case. Obviously,
-// this function requires a seekable file or stream and leaves the file
-// pointer undefined. A return value of -1 indicates the length could not
-// be determined.
+// Return the file format specified in the call to WavpackSetFileInformation()
+// when the file was created. For all files created prior to WavPack 5.0 this
+// will 0 (WP_FORMAT_WAV).
 
-static uint32_t seek_final_index (WavpackStreamReader *reader, void *id)
+unsigned char WavpackGetFileFormat (WavpackContext *wpc)
 {
-    uint32_t result = (uint32_t) -1, bcount;
-    WavpackHeader wphdr;
-    unsigned char *tempbuff;
+    return wpc->file_format;
+}
 
-    if (reader->get_length (id) > 1200000L)
-        reader->set_pos_rel (id, -1048576L, SEEK_END);
+// Return a string representing the recommended file extension for the open
+// WavPack file. For all files created prior to WavPack 5.0 this will be "wav",
+// even for raw files with no RIFF into. This string is specified in the
+// call to WavpackSetFileInformation() when the file was created.
+
+char *WavpackGetFileExtension (WavpackContext *wpc)
+{
+    if (wpc && wpc->file_extension [0])
+        return wpc->file_extension;
     else
-        reader->set_pos_abs (id, 0);
-
-    while (1) {
-        bcount = read_next_header (reader, id, &wphdr);
-
-        if (bcount == (uint32_t) -1)
-            return result;
-
-        tempbuff = malloc (wphdr.ckSize + 8);
-        if (!tempbuff)
-            return result;
-        memcpy (tempbuff, &wphdr, 32);
-
-        if (reader->read_bytes (id, tempbuff + 32, wphdr.ckSize - 24) != wphdr.ckSize - 24) {
-            free (tempbuff);
-            return result;
-        }
-
-        free (tempbuff);
-
-        if (wphdr.block_samples && (wphdr.flags & FINAL_BLOCK))
-            result = wphdr.block_index + wphdr.block_samples;
-    }
+        return "wav";
 }
 
 // This function initializes everything required to unpack a WavPack block
@@ -265,6 +293,7 @@ int unpack_init (WavpackContext *wpc)
     wps->num_terms = 0;
     wps->mute_error = FALSE;
     wps->crc = wps->crc_x = 0xffffffff;
+    wps->dsd.ready = 0;
     CLEAR (wps->wvbits);
     CLEAR (wps->wvcbits);
     CLEAR (wps->wvxbits);
@@ -301,7 +330,7 @@ int unpack_init (WavpackContext *wpc)
             }
     }
 
-    if (wps->wphdr.block_samples && !bs_is_open (&wps->wvbits)) {
+    if (wps->wphdr.block_samples && ((wps->wphdr.flags & DSD_FLAG) ? !wps->dsd.ready : !bs_is_open (&wps->wvbits))) {
         if (bs_is_open (&wps->wvcbits))
             strcpy (wpc->error_message, "can't unpack correction files alone!");
 
@@ -319,7 +348,7 @@ int unpack_init (WavpackContext *wpc)
     }
 
     if (wps->wphdr.block_samples)
-        wps->sample_index = wps->wphdr.block_index;
+        wps->sample_index = GET_BLOCK_INDEX (wps->wphdr);
 
     return TRUE;
 }
@@ -334,7 +363,7 @@ int unpack_init (WavpackContext *wpc)
 
 static int init_wv_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
-    if (!wpmd->byte_length)
+    if (!wpmd->byte_length || (wpmd->byte_length & 1))
         return FALSE;
 
     bs_open_read (&wps->wvbits, wpmd->data, (unsigned char *) wpmd->data + wpmd->byte_length);
@@ -346,7 +375,7 @@ static int init_wv_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 
 static int init_wvc_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
-    if (!wpmd->byte_length)
+    if (!wpmd->byte_length || (wpmd->byte_length & 1))
         return FALSE;
 
     bs_open_read (&wps->wvcbits, wpmd->data, (unsigned char *) wpmd->data + wpmd->byte_length);
@@ -364,7 +393,7 @@ static int init_wvx_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
     unsigned char *cp = wpmd->data;
 
-    if (wpmd->byte_length <= 4)
+    if (wpmd->byte_length <= 4 || (wpmd->byte_length & 1))
         return FALSE;
 
     wps->crc_wvx = *cp++;
@@ -421,12 +450,14 @@ static int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
     unsigned char *byteptr = wpmd->data;
     uint32_t mask = 0;
 
-    if (!bytecnt || bytecnt > 6)
+    if (!bytecnt || bytecnt > 7)
         return FALSE;
 
     if (!wpc->config.num_channels) {
 
-        if (bytecnt == 6) {
+        // if bytecnt is 6 or 7 we are using new configuration with "unlimited" streams
+
+        if (bytecnt >= 6) {
             wpc->config.num_channels = (byteptr [0] | ((byteptr [2] & 0xf) << 8)) + 1;
             wpc->max_streams = (byteptr [1] | ((byteptr [2] & 0xf0) << 4)) + 1;
 
@@ -436,7 +467,10 @@ static int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
             byteptr += 3;
             mask = *byteptr++;
             mask |= (uint32_t) *byteptr++ << 8;
-            mask |= (uint32_t) *byteptr << 16;
+            mask |= (uint32_t) *byteptr++ << 16;
+
+            if (bytecnt == 7)                           // this was introduced in 5.0
+                mask |= (uint32_t) *byteptr << 24;
         }
         else {
             wpc->config.num_channels = *byteptr++;
@@ -460,6 +494,23 @@ static int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
     return TRUE;
 }
 
+// Read multichannel identity information from metadata. Data is an array of
+// unsigned characters representing any channels in the file that DO NOT
+// match one the 18 Microsoft standard channels (and are represented in the
+// channel mask). A value of 0 is not allowed and 0xff means an unknown or
+// undefined channel identity.
+
+static int read_channel_identities (WavpackContext *wpc, WavpackMetadata *wpmd)
+{
+    if (!wpc->channel_identities) {
+        wpc->channel_identities = malloc (wpmd->byte_length + 1);
+        memcpy (wpc->channel_identities, wpmd->data, wpmd->byte_length);
+        wpc->channel_identities [wpmd->byte_length] = 0;
+    }
+
+    return TRUE;
+}
+
 // Read configuration information from metadata.
 
 static int read_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
@@ -472,9 +523,86 @@ static int read_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
         wpc->config.flags |= (int32_t) *byteptr++ << 8;
         wpc->config.flags |= (int32_t) *byteptr++ << 16;
         wpc->config.flags |= (int32_t) *byteptr++ << 24;
+        bytecnt -= 3;
 
-        if (bytecnt >= 4 && (wpc->config.flags & CONFIG_EXTRA_MODE))
-            wpc->config.xmode = *byteptr;
+        if (bytecnt && (wpc->config.flags & CONFIG_EXTRA_MODE)) {
+            wpc->config.xmode = *byteptr++;
+            bytecnt--;
+        }
+
+        // we used an extra config byte here for the 5.0.0 alpha, so still
+        // honor it now (but this has been replaced with NEW_CONFIG)
+
+        if (bytecnt) {
+            wpc->config.qmode = (wpc->config.qmode & ~0xff) | *byteptr;
+            wpc->version_five = 1;
+        }
+    }
+
+    return TRUE;
+}
+
+// Read "new" configuration information from metadata.
+
+static int read_new_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
+{
+    int bytecnt = wpmd->byte_length;
+    unsigned char *byteptr = wpmd->data;
+
+    wpc->version_five = 1;      // just having this block signals version 5.0
+
+    wpc->file_format = wpc->config.qmode = wpc->channel_layout = 0;
+
+    if (wpc->channel_reordering) {
+        free (wpc->channel_reordering);
+        wpc->channel_reordering = NULL;
+    }
+
+    // if there's any data, the first two bytes are file_format and qmode flags
+
+    if (bytecnt) {
+        wpc->file_format = *byteptr++;
+        wpc->config.qmode = (wpc->config.qmode & ~0xff) | *byteptr++;
+        bytecnt -= 2;
+
+        // another byte indicates a channel layout
+
+        if (bytecnt) {
+            int nchans, i;
+
+            wpc->channel_layout = (int32_t) *byteptr++ << 16;
+            bytecnt--;
+
+            // another byte means we have a channel count for the layout and maybe a reordering
+
+            if (bytecnt) {
+                wpc->channel_layout += nchans = *byteptr++;
+                bytecnt--;
+
+                // any more means there's a reordering string
+
+                if (bytecnt) {
+                    if (bytecnt > nchans)
+                        return FALSE;
+
+                    wpc->channel_reordering = malloc (nchans);
+
+                    // note that redundant reordering info is not stored, so we fill in the rest
+
+                    if (wpc->channel_reordering) {
+                        for (i = 0; i < nchans; ++i)
+                            if (bytecnt) {
+                                wpc->channel_reordering [i] = *byteptr++;
+                                bytecnt--;
+                            }
+                            else
+                                wpc->channel_reordering [i] = i;
+                    }
+                }
+            }
+            else
+                wpc->channel_layout += wpc->config.num_channels;
+        }
     }
 
     return TRUE;
@@ -487,10 +615,15 @@ static int read_sample_rate (WavpackContext *wpc, WavpackMetadata *wpmd)
     int bytecnt = wpmd->byte_length;
     unsigned char *byteptr = wpmd->data;
 
-    if (bytecnt == 3) {
+    if (bytecnt == 3 || bytecnt == 4) {
         wpc->config.sample_rate = (int32_t) *byteptr++;
         wpc->config.sample_rate |= (int32_t) *byteptr++ << 8;
         wpc->config.sample_rate |= (int32_t) *byteptr++ << 16;
+
+        // for sampling rates > 16777215 (non-audio probably, or ...)
+
+        if (bytecnt == 4)
+            wpc->config.sample_rate |= (int32_t) (*byteptr & 0x7f) << 24;
     }
 
     return TRUE;
@@ -593,9 +726,15 @@ static int process_metadata (WavpackContext *wpc, WavpackMetadata *wpmd)
         case ID_CHANNEL_INFO:
             return read_channel_info (wpc, wpmd);
 
+        case ID_CHANNEL_IDENTITIES:
+            return read_channel_identities (wpc, wpmd);
+
         case ID_CONFIG_BLOCK:
             return read_config_info (wpc, wpmd);
 
+        case ID_NEW_CONFIG_BLOCK:
+            return read_new_config_info (wpc, wpmd);
+
         case ID_SAMPLE_RATE:
             return read_sample_rate (wpc, wpmd);
 
@@ -608,9 +747,25 @@ static int process_metadata (WavpackContext *wpc, WavpackMetadata *wpmd)
         case ID_WVX_BITSTREAM:
             return init_wvx_bitstream (wps, wpmd);
 
+        case ID_DSD_BLOCK:
+#ifdef ENABLE_DSD
+            return init_dsd_block (wpc, wpmd);
+#else
+            strcpy (wpc->error_message, "not configured to handle DSD WavPack files!");
+            return FALSE;
+#endif
+
+        case ID_ALT_HEADER: case ID_ALT_TRAILER:
+            if (!(wpc->open_flags & OPEN_ALT_TYPES))
+                return TRUE;
+
         case ID_RIFF_HEADER: case ID_RIFF_TRAILER:
             return read_wrapper_data (wpc, wpmd);
 
+        case ID_ALT_MD5_CHECKSUM:
+            if (!(wpc->open_flags & OPEN_ALT_TYPES))
+                return TRUE;
+
         case ID_MD5_CHECKSUM:
             if (wpmd->byte_length == 16) {
                 memcpy (wpc->config.md5_checksum, wpmd->data, 16);
@@ -620,6 +775,21 @@ static int process_metadata (WavpackContext *wpc, WavpackMetadata *wpmd)
 
             return TRUE;
 
+        case ID_ALT_EXTENSION:
+            if (wpmd->byte_length && wpmd->byte_length < sizeof (wpc->file_extension)) {
+                memcpy (wpc->file_extension, wpmd->data, wpmd->byte_length);
+                wpc->file_extension [wpmd->byte_length] = 0;
+            }
+
+            return TRUE;
+
+        // we don't actually verify the checksum here (it's done right after the
+        // block is read), but it's a good indicator of version 5 files
+
+        case ID_BLOCK_CHECKSUM:
+            wpc->version_five = 1;
+            return TRUE;
+
         default:
             return (wpmd->id & ID_OPTIONAL_DATA) ? TRUE : FALSE;
     }
@@ -646,7 +816,7 @@ static void bs_open_read (Bitstream *bs, void *buffer_start, void *buffer_end)
 
 static void bs_read (Bitstream *bs)
 {
-    bs->ptr = bs->buf - 1;
+    bs->ptr = bs->buf;
     bs->error = 1;
 }
 
@@ -676,45 +846,27 @@ uint32_t bs_close_read (Bitstream *bs)
 // be used for seekable files (not pipes) and is not available for pre-4.0 WavPack
 // files.
 
-static int seek_riff_trailer (WavpackContext *wpc);
-
 void WavpackSeekTrailingWrapper (WavpackContext *wpc)
 {
     if ((wpc->open_flags & OPEN_WRAPPER) &&
-        wpc->reader->can_seek (wpc->wv_in) && !wpc->stream3) {
-            uint32_t pos_save = wpc->reader->get_pos (wpc->wv_in);
-
-            seek_riff_trailer (wpc);
-            wpc->reader->set_pos_abs (wpc->wv_in, pos_save);
-    }
+        wpc->reader->can_seek (wpc->wv_in) && !wpc->stream3)
+            seek_eof_information (wpc, NULL, TRUE);
 }
 
 // Get any MD5 checksum stored in the metadata (should be called after reading
 // last sample or an extra seek will occur). A return value of FALSE indicates
 // that no MD5 checksum was stored.
 
-static int seek_md5 (WavpackStreamReader *reader, void *id, unsigned char data [16]);
-
 int WavpackGetMD5Sum (WavpackContext *wpc, unsigned char data [16])
 {
     if (wpc->config.flags & CONFIG_MD5_CHECKSUM) {
+        if (!wpc->config.md5_read && wpc->reader->can_seek (wpc->wv_in))
+            seek_eof_information (wpc, NULL, FALSE);
+
         if (wpc->config.md5_read) {
             memcpy (data, wpc->config.md5_checksum, 16);
             return TRUE;
         }
-        else if (wpc->reader->can_seek (wpc->wv_in)) {
-            uint32_t pos_save = wpc->reader->get_pos (wpc->wv_in);
-
-            wpc->config.md5_read = seek_md5 (wpc->reader, wpc->wv_in, wpc->config.md5_checksum);
-            wpc->reader->set_pos_abs (wpc->wv_in, pos_save);
-
-            if (wpc->config.md5_read) {
-                memcpy (data, wpc->config.md5_checksum, 16);
-                return TRUE;
-            }
-            else
-                return FALSE;
-        }
     }
 
     return FALSE;
@@ -726,7 +878,7 @@ int WavpackGetMD5Sum (WavpackContext *wpc, unsigned char data [16])
 // to indicate the error. No additional bytes are read past the header and it
 // is returned in the processor's native endian mode. Seeking is not required.
 
-uint32_t read_next_header (WavpackStreamReader *reader, void *id, WavpackHeader *wphdr)
+uint32_t read_next_header (WavpackStreamReader64 *reader, void *id, WavpackHeader *wphdr)
 {
     unsigned char buffer [sizeof (*wphdr)], *sp = buffer + sizeof (*wphdr), *ep = sp;
     uint32_t bytes_skipped = 0;
@@ -771,7 +923,7 @@ uint32_t read_next_header (WavpackStreamReader *reader, void *id, WavpackHeader
 
 static int match_wvc_header (WavpackHeader *wv_hdr, WavpackHeader *wvc_hdr)
 {
-    if (wv_hdr->block_index == wvc_hdr->block_index &&
+    if (GET_BLOCK_INDEX (*wv_hdr) == GET_BLOCK_INDEX (*wvc_hdr) &&
         wv_hdr->block_samples == wvc_hdr->block_samples) {
             int wvi = 0, wvci = 0;
 
@@ -793,7 +945,7 @@ static int match_wvc_header (WavpackHeader *wv_hdr, WavpackHeader *wvc_hdr)
             return (wvci - wvi < 0) ? 1 : -1;
         }
 
-    if ((int32_t)(wvc_hdr->block_index - wv_hdr->block_index) < 0)
+    if (((GET_BLOCK_INDEX (*wvc_hdr) - GET_BLOCK_INDEX (*wv_hdr)) << 24) < 0)
         return 1;
     else
         return -1;
@@ -810,7 +962,8 @@ static int match_wvc_header (WavpackHeader *wv_hdr, WavpackHeader *wvc_hdr)
 int read_wvc_block (WavpackContext *wpc)
 {
     WavpackStream *wps = wpc->streams [wpc->current_stream];
-    uint32_t bcount, file2pos;
+    int64_t bcount, file2pos;
+    WavpackHeader orig_wphdr;
     WavpackHeader wphdr;
     int compare_result;
 
@@ -824,10 +977,12 @@ int read_wvc_block (WavpackContext *wpc)
             return FALSE;
         }
 
+        memcpy (&orig_wphdr, &wphdr, 32);       // save original header for verify step
+
         if (wpc->open_flags & OPEN_STREAMING)
-            wphdr.block_index = wps->sample_index = 0;
+            SET_BLOCK_INDEX (wphdr, wps->sample_index = 0);
         else
-            wphdr.block_index -= wpc->initial_index;
+            SET_BLOCK_INDEX (wphdr, GET_BLOCK_INDEX (wphdr) - wpc->initial_index);
 
         if (wphdr.flags & INITIAL_BLOCK)
             wpc->file2pos = file2pos + bcount;
@@ -838,10 +993,9 @@ int read_wvc_block (WavpackContext *wpc)
             wps->block2buff = malloc (wphdr.ckSize + 8);
 	    if (!wps->block2buff)
 	        return FALSE;
-            memcpy (wps->block2buff, &wphdr, 32);
 
             if (wpc->reader->read_bytes (wpc->wvc_in, wps->block2buff + 32, wphdr.ckSize - 24) !=
-                wphdr.ckSize - 24 || (wphdr.flags & UNKNOWN_FLAGS)) {
+                wphdr.ckSize - 24) {
                     free (wps->block2buff);
                     wps->block2buff = NULL;
                     wps->wvc_skip = TRUE;
@@ -849,7 +1003,19 @@ int read_wvc_block (WavpackContext *wpc)
                     return FALSE;
             }
 
+            memcpy (wps->block2buff, &orig_wphdr, 32);
+
+            // don't use corrupt blocks
+            if (!WavpackVerifySingleBlock (wps->block2buff, !(wpc->open_flags & OPEN_NO_CHECKSUM))) {
+                free (wps->block2buff);
+                wps->block2buff = NULL;
+                wps->wvc_skip = TRUE;
+                wpc->crc_errors++;
+                return TRUE;
+            }
+
             wps->wvc_skip = FALSE;
+            memcpy (wps->block2buff, &wphdr, 32);
             memcpy (&wps->wphdr, &wphdr, 32);
             return TRUE;
         }
@@ -862,95 +1028,157 @@ int read_wvc_block (WavpackContext *wpc)
     }
 }
 
-static int seek_md5 (WavpackStreamReader *reader, void *id, unsigned char data [16])
+// This function is used to seek to end of a file to obtain certain information
+// that is stored there at the file creation time because it is not known at
+// the start. This includes the MD5 sum and and trailing part of the file
+// wrapper, and in some rare cases may include the total number of samples in
+// the file (although we usually try to back up and write that at the front of
+// the file). Note this function restores the file position to its original
+// location (and obviously requires a seekable file). The normal return value
+// is TRUE indicating no errors, although this does not actually mean that any
+// information was retrieved. An error return of FALSE usually means the file
+// terminated unexpectedly. Note that this could be used to get all three
+// types of information in one go, but it's not actually used that way now.
+
+static int seek_eof_information (WavpackContext *wpc, int64_t *final_index, int get_wrapper)
 {
-    unsigned char meta_id, c1, c2;
-    uint32_t bcount, meta_bc;
-    WavpackHeader wphdr;
-
-    if (reader->get_length (id) > 1200000L)
-        reader->set_pos_rel (id, -1048576L, SEEK_END);
-
-    while (1) {
-        bcount = read_next_header (reader, id, &wphdr);
-
-        if (bcount == (uint32_t) -1)
-            return FALSE;
-
-        bcount = wphdr.ckSize - sizeof (WavpackHeader) + 8;
-
-        while (bcount >= 2) {
-            if (reader->read_bytes (id, &meta_id, 1) != 1 ||
-                reader->read_bytes (id, &c1, 1) != 1)
-                    return FALSE;
-
-            meta_bc = c1 << 1;
-            bcount -= 2;
-
-            if (meta_id & ID_LARGE) {
-                if (bcount < 2 || reader->read_bytes (id, &c1, 1) != 1 ||
-                    reader->read_bytes (id, &c2, 1) != 1)
-                        return FALSE;
-
-                meta_bc += ((uint32_t) c1 << 9) + ((uint32_t) c2 << 17);
-                bcount -= 2;
-            }
-
-            if (meta_id == ID_MD5_CHECKSUM)
-                return (meta_bc == 16 && bcount >= 16 &&
-                    reader->read_bytes (id, data, 16) == 16);
-
-            reader->set_pos_rel (id, meta_bc, SEEK_CUR);
-            bcount -= meta_bc;
-        }
-    }
-}
-
-static int seek_riff_trailer (WavpackContext *wpc)
-{
-    WavpackStreamReader *reader = wpc->reader;
+    int64_t restore_pos, last_pos = -1;
+    WavpackStreamReader64 *reader = wpc->reader;
+    int alt_types = wpc->open_flags & OPEN_ALT_TYPES;
+    uint32_t blocks = 0, audio_blocks = 0;
     void *id = wpc->wv_in;
-    unsigned char meta_id, c1, c2;
-    uint32_t bcount, meta_bc;
     WavpackHeader wphdr;
 
-    if (reader->get_length (id) > 1200000L)
-        reader->set_pos_rel (id, -1048576L, SEEK_END);
+    restore_pos = reader->get_pos (id);    // we restore file position when done
+
+    // start 1MB from the end-of-file, or from the start if the file is not that big
+
+    if (reader->get_length (id) > 1048576LL)
+        reader->set_pos_rel (id, -1048576, SEEK_END);
+    else
+        reader->set_pos_abs (id, 0);
+
+    // Note that we go backward (without parsing inside blocks) until we find a block
+    // with audio (careful to not get stuck in a loop). Only then do we go forward
+    // parsing all blocks in their entirety.
 
     while (1) {
-        bcount = read_next_header (reader, id, &wphdr);
+        uint32_t bcount = read_next_header (reader, id, &wphdr);
+        int64_t current_pos = reader->get_pos (id);
 
-        if (bcount == (uint32_t) -1)
+        // if we just got to the same place as last time, we're stuck and need to give up
+
+        if (current_pos == last_pos) {
+            reader->set_pos_abs (id, restore_pos);
+            return FALSE;
+        }
+
+        last_pos = current_pos;
+
+        // We enter here if we just read 1 MB without seeing any WavPack block headers.
+        // Since WavPack blocks are < 1 MB, that means we're in a big APE tag, or we got
+        // to the end-of-file.
+
+        if (bcount == (uint32_t) -1) {
+
+            // if we have not seen any blocks at all yet, back up almost 2 MB (or to the
+            // beginning of the file) and try again
+
+            if (!blocks) {
+                if (current_pos > 2000000LL)
+                    reader->set_pos_rel (id, -2000000, SEEK_CUR);
+                else
+                    reader->set_pos_abs (id, 0);
+
+                continue;
+            }
+
+            // if we have seen WavPack blocks, then this means we've done all we can do here
+
+            reader->set_pos_abs (id, restore_pos);
             return TRUE;
+        }
+
+        blocks++;
+
+        // If the block has audio samples, calculate a final index, although this is not
+        // final since this may not be the last block with audio. On the other hand, if
+        // this block does not have audio, and we haven't seen one with audio, we have
+        // to go back some more.
+
+        if (wphdr.block_samples) {
+            if (final_index)
+                *final_index = GET_BLOCK_INDEX (wphdr) + wphdr.block_samples;
+
+            audio_blocks++;
+        }
+        else if (!audio_blocks) {
+            if (current_pos > 1048576LL)
+                reader->set_pos_rel (id, -1048576, SEEK_CUR);
+            else
+                reader->set_pos_abs (id, 0);
+
+            continue;
+        }
+
+        // at this point we have seen at least one block with audio, so we parse the
+        // entire block looking for MD5 metadata or (conditionally) trailing wrappers
 
         bcount = wphdr.ckSize - sizeof (WavpackHeader) + 8;
 
         while (bcount >= 2) {
+            unsigned char meta_id, c1, c2;
+            uint32_t meta_bc, meta_size;
+
             if (reader->read_bytes (id, &meta_id, 1) != 1 ||
-                reader->read_bytes (id, &c1, 1) != 1)
-                    return TRUE;
+                reader->read_bytes (id, &c1, 1) != 1) {
+                    reader->set_pos_abs (id, restore_pos);
+                    return FALSE;
+            }
 
             meta_bc = c1 << 1;
             bcount -= 2;
 
             if (meta_id & ID_LARGE) {
                 if (bcount < 2 || reader->read_bytes (id, &c1, 1) != 1 ||
-                    reader->read_bytes (id, &c2, 1) != 1)
-                        return TRUE;
+                    reader->read_bytes (id, &c2, 1) != 1) {
+                        reader->set_pos_abs (id, restore_pos);
+                        return FALSE;
+                }
 
                 meta_bc += ((uint32_t) c1 << 9) + ((uint32_t) c2 << 17);
                 bcount -= 2;
             }
 
-            if ((meta_id & ID_UNIQUE) == ID_RIFF_TRAILER && meta_bc) {
+            meta_size = (meta_id & ID_ODD_SIZE) ? meta_bc - 1 : meta_bc;
+            meta_id &= ID_UNIQUE;
+
+            if (get_wrapper && (meta_id == ID_RIFF_TRAILER || (alt_types && meta_id == ID_ALT_TRAILER)) && meta_bc) {
                 wpc->wrapper_data = realloc (wpc->wrapper_data, wpc->wrapper_bytes + meta_bc);
-		if (!wpc->wrapper_data)
-		    return FALSE;
+
+                if (!wpc->wrapper_data) {
+                    reader->set_pos_abs (id, restore_pos);
+                    return FALSE;
+                }
 
                 if (reader->read_bytes (id, wpc->wrapper_data + wpc->wrapper_bytes, meta_bc) == meta_bc)
-                    wpc->wrapper_bytes += meta_bc;
+                    wpc->wrapper_bytes += meta_size;
+                else {
+                    reader->set_pos_abs (id, restore_pos);
+                    return FALSE;
+                }
+            }
+            else if (meta_id == ID_MD5_CHECKSUM || (alt_types && meta_id == ID_ALT_MD5_CHECKSUM)) {
+                if (meta_bc == 16 && bcount >= 16) {
+                    if (reader->read_bytes (id, wpc->config.md5_checksum, 16) == 16)
+                        wpc->config.md5_read = TRUE;
+                    else {
+                        reader->set_pos_abs (id, restore_pos);
+                        return FALSE;
+                    }
+                }
                 else
-                    return TRUE;
+                    reader->set_pos_rel (id, meta_bc, SEEK_CUR);
             }
             else
                 reader->set_pos_rel (id, meta_bc, SEEK_CUR);
@@ -959,3 +1187,89 @@ static int seek_riff_trailer (WavpackContext *wpc)
         }
     }
 }
+
+// Quickly verify the referenced block. It is assumed that the WavPack header has been converted
+// to native endian format. If a block checksum is performed, that is done in little-endian
+// (file) format. It is also assumed that the caller has made sure that the block length
+// indicated in the header is correct (we won't overflow the buffer). If a checksum is present,
+// then it is checked, otherwise we just check that all the metadata blocks are formatted
+// correctly (without looking at their contents). Returns FALSE for bad block.
+
+int WavpackVerifySingleBlock (unsigned char *buffer, int verify_checksum)
+{
+    WavpackHeader *wphdr = (WavpackHeader *) buffer;
+    uint32_t checksum_passed = 0, bcount, meta_bc;
+    unsigned char *dp, meta_id, c1, c2;
+
+    if (strncmp (wphdr->ckID, "wvpk", 4) || wphdr->ckSize + 8 < sizeof (WavpackHeader))
+        return FALSE;
+
+    bcount = wphdr->ckSize - sizeof (WavpackHeader) + 8;
+    dp = (unsigned char *)(wphdr + 1);
+
+    while (bcount >= 2) {
+        meta_id = *dp++;
+        c1 = *dp++;
+
+        meta_bc = c1 << 1;
+        bcount -= 2;
+
+        if (meta_id & ID_LARGE) {
+            if (bcount < 2)
+                return FALSE;
+
+            c1 = *dp++;
+            c2 = *dp++;
+            meta_bc += ((uint32_t) c1 << 9) + ((uint32_t) c2 << 17);
+            bcount -= 2;
+        }
+
+        if (bcount < meta_bc)
+            return FALSE;
+
+        if (verify_checksum && (meta_id & ID_UNIQUE) == ID_BLOCK_CHECKSUM) {
+#ifdef BITSTREAM_SHORTS
+            uint16_t *csptr = (uint16_t*) buffer;
+#else
+            unsigned char *csptr = buffer;
+#endif
+            int wcount = (int)(dp - 2 - buffer) >> 1;
+            uint32_t csum = (uint32_t) -1;
+
+            if ((meta_id & ID_ODD_SIZE) || meta_bc < 2 || meta_bc > 4)
+                return FALSE;
+
+#ifdef BITSTREAM_SHORTS
+            while (wcount--)
+                csum = (csum * 3) + *csptr++;
+#else
+            WavpackNativeToLittleEndian ((WavpackHeader *) buffer, WavpackHeaderFormat);
+
+            while (wcount--) {
+                csum = (csum * 3) + csptr [0] + (csptr [1] << 8);
+                csptr += 2;
+            }
+
+            WavpackLittleEndianToNative ((WavpackHeader *) buffer, WavpackHeaderFormat);
+#endif
+
+            if (meta_bc == 4) {
+                if (*dp++ != (csum & 0xff) || *dp++ != ((csum >> 8) & 0xff) || *dp++ != ((csum >> 16) & 0xff) || *dp++ != ((csum >> 24) & 0xff))
+                    return FALSE;
+            }
+            else {
+                csum ^= csum >> 16;
+
+                if (*dp++ != (csum & 0xff) || *dp++ != ((csum >> 8) & 0xff))
+                    return FALSE;
+            }
+
+            checksum_passed++;
+        }
+
+        bcount -= meta_bc;
+        dp += meta_bc;
+    }
+
+    return (bcount == 0) && (!verify_checksum || !(wphdr->flags & HAS_CHECKSUM) || checksum_passed);
+}
