@@ -173,27 +173,27 @@ static void write_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
     for (dpp = wps->decorr_passes; tcount--; ++dpp)
         if (wcount) {
             if (dpp->term > MAX_TERM) {
-                dpp->samples_A [0] = exp2s (temp = log2s (dpp->samples_A [0]));
+                dpp->samples_A [0] = wp_exp2s (temp = wp_log2s (dpp->samples_A [0]));
                 *byteptr++ = temp;
                 *byteptr++ = temp >> 8;
-                dpp->samples_A [1] = exp2s (temp = log2s (dpp->samples_A [1]));
+                dpp->samples_A [1] = wp_exp2s (temp = wp_log2s (dpp->samples_A [1]));
                 *byteptr++ = temp;
                 *byteptr++ = temp >> 8;
 
                 if (!(wps->wphdr.flags & MONO_DATA)) {
-                    dpp->samples_B [0] = exp2s (temp = log2s (dpp->samples_B [0]));
+                    dpp->samples_B [0] = wp_exp2s (temp = wp_log2s (dpp->samples_B [0]));
                     *byteptr++ = temp;
                     *byteptr++ = temp >> 8;
-                    dpp->samples_B [1] = exp2s (temp = log2s (dpp->samples_B [1]));
+                    dpp->samples_B [1] = wp_exp2s (temp = wp_log2s (dpp->samples_B [1]));
                     *byteptr++ = temp;
                     *byteptr++ = temp >> 8;
                 }
             }
             else if (dpp->term < 0) {
-                dpp->samples_A [0] = exp2s (temp = log2s (dpp->samples_A [0]));
+                dpp->samples_A [0] = wp_exp2s (temp = wp_log2s (dpp->samples_A [0]));
                 *byteptr++ = temp;
                 *byteptr++ = temp >> 8;
-                dpp->samples_B [0] = exp2s (temp = log2s (dpp->samples_B [0]));
+                dpp->samples_B [0] = wp_exp2s (temp = wp_log2s (dpp->samples_B [0]));
                 *byteptr++ = temp;
                 *byteptr++ = temp >> 8;
             }
@@ -201,12 +201,12 @@ static void write_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
                 int m = 0, cnt = dpp->term;
 
                 while (cnt--) {
-                    dpp->samples_A [m] = exp2s (temp = log2s (dpp->samples_A [m]));
+                    dpp->samples_A [m] = wp_exp2s (temp = wp_log2s (dpp->samples_A [m]));
                     *byteptr++ = temp;
                     *byteptr++ = temp >> 8;
 
                     if (!(wps->wphdr.flags & MONO_DATA)) {
-                        dpp->samples_B [m] = exp2s (temp = log2s (dpp->samples_B [m]));
+                        dpp->samples_B [m] = wp_exp2s (temp = wp_log2s (dpp->samples_B [m]));
                         *byteptr++ = temp;
                         *byteptr++ = temp >> 8;
                     }
@@ -239,29 +239,29 @@ static void write_shaping_info (WavpackStream *wps, WavpackMetadata *wpmd)
     byteptr = wpmd->data = malloc (12);
     wpmd->id = ID_SHAPING_WEIGHTS;
 
-    wps->dc.error [0] = exp2s (temp = log2s (wps->dc.error [0]));
+    wps->dc.error [0] = wp_exp2s (temp = wp_log2s (wps->dc.error [0]));
     *byteptr++ = temp;
     *byteptr++ = temp >> 8;
-    wps->dc.shaping_acc [0] = exp2s (temp = log2s (wps->dc.shaping_acc [0]));
+    wps->dc.shaping_acc [0] = wp_exp2s (temp = wp_log2s (wps->dc.shaping_acc [0]));
     *byteptr++ = temp;
     *byteptr++ = temp >> 8;
 
     if (!(wps->wphdr.flags & MONO_DATA)) {
-        wps->dc.error [1] = exp2s (temp = log2s (wps->dc.error [1]));
+        wps->dc.error [1] = wp_exp2s (temp = wp_log2s (wps->dc.error [1]));
         *byteptr++ = temp;
         *byteptr++ = temp >> 8;
-        wps->dc.shaping_acc [1] = exp2s (temp = log2s (wps->dc.shaping_acc [1]));
+        wps->dc.shaping_acc [1] = wp_exp2s (temp = wp_log2s (wps->dc.shaping_acc [1]));
         *byteptr++ = temp;
         *byteptr++ = temp >> 8;
     }
 
     if (wps->dc.shaping_delta [0] | wps->dc.shaping_delta [1]) {
-        wps->dc.shaping_delta [0] = exp2s (temp = log2s (wps->dc.shaping_delta [0]));
+        wps->dc.shaping_delta [0] = wp_exp2s (temp = wp_log2s (wps->dc.shaping_delta [0]));
         *byteptr++ = temp;
         *byteptr++ = temp >> 8;
 
         if (!(wps->wphdr.flags & MONO_DATA)) {
-            wps->dc.shaping_delta [1] = exp2s (temp = log2s (wps->dc.shaping_delta [1]));
+            wps->dc.shaping_delta [1] = wp_exp2s (temp = wp_log2s (wps->dc.shaping_delta [1]));
             *byteptr++ = temp;
             *byteptr++ = temp >> 8;
         }
@@ -309,21 +309,22 @@ static void write_float_info (WavpackStream *wps, WavpackMetadata *wpmd)
 static void write_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
 {
     uint32_t mask = wpc->config.channel_mask;
-    char *byteptr;
+    char *byteptr = wpmd->data = malloc (8);
 
-    if (wpc->num_streams > OLD_MAX_STREAMS) {
-        byteptr = wpmd->data = malloc (6);
-        wpmd->id = ID_CHANNEL_INFO;
-        *byteptr++ = wpc->config.num_channels - 1;
-        *byteptr++ = wpc->num_streams - 1;
+    wpmd->id = ID_CHANNEL_INFO;
+
+    if (wpc->num_streams > OLD_MAX_STREAMS) {       // if > 8 streams, use 6 or 7 bytes (breaks old decoders
+        *byteptr++ = wpc->config.num_channels - 1;  // that could only handle 8 streams) and allow (in theory)
+        *byteptr++ = wpc->num_streams - 1;          // up to 4096 channels
         *byteptr++ = (((wpc->num_streams - 1) >> 4) & 0xf0) | (((wpc->config.num_channels - 1) >> 8) & 0xf);
         *byteptr++ = mask;
         *byteptr++ = (mask >> 8);
         *byteptr++ = (mask >> 16);
+
+        if (mask & 0xff000000)                      // this will break versions < 5.0, but is RF64-specific
+            *byteptr++ = (mask >> 24);
     }
-    else {
-        byteptr = wpmd->data = malloc (4);
-        wpmd->id = ID_CHANNEL_INFO;
+    else {                                          // otherwise use only 1 to 5 bytes
         *byteptr++ = wpc->config.num_channels;
 
         while (mask) {
@@ -335,6 +336,19 @@ static void write_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
     wpmd->byte_length = (int32_t)(byteptr - (char *) wpmd->data);
 }
 
+// Allocate room for and copy the multichannel identities into the specified
+// metadata structure. Data is an array of unsigned characters representing
+// any channels in the file that DO NOT match one the 18 Microsoft standard
+// channels (and are represented in the channel mask). A value of 0 is not
+// allowed and 0xff means an unknown or undefined channel identity.
+
+static void write_channel_identities_info (WavpackContext *wpc, WavpackMetadata *wpmd)
+{
+    wpmd->byte_length = (int) strlen ((char *) wpc->channel_identities);
+    wpmd->data = strdup ((char *) wpc->channel_identities);
+    wpmd->id = ID_CHANNEL_IDENTITIES;
+}
+
 // Allocate room for and copy the configuration information into the specified
 // metadata structure. Currently, we just store the upper 3 bytes of
 // config.flags and only in the first block of audio data. Note that this is
@@ -345,7 +359,7 @@ static void write_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
 {
     char *byteptr;
 
-    byteptr = wpmd->data = malloc (4);
+    byteptr = wpmd->data = malloc (8);
     wpmd->id = ID_CONFIG_BLOCK;
     *byteptr++ = (char) (wpc->config.flags >> 8);
     *byteptr++ = (char) (wpc->config.flags >> 16);
@@ -354,13 +368,62 @@ static void write_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
     if (wpc->config.flags & CONFIG_EXTRA_MODE)
         *byteptr++ = (char) wpc->config.xmode;
 
+    // for the 5.0.0 alpha, we wrote the qmode flags here, but this
+    // has been replaced with the new_config block
+    // *byteptr++ = (char) wpc->config.qmode;
+
     wpmd->byte_length = (int32_t)(byteptr - (char *) wpmd->data);
 }
 
-// Allocate room for and copy the non-standard sampling rateinto the specified
-// metadata structure. We just store the lower 3 bytes of the sampling rate.
-// Note that this would only be used when the sampling rate was not included
-// in the table of 15 "standard" values.
+// Allocate room for and copy the "new" configuration information into the
+// specified metadata structure. This is all the stuff introduced with version
+// 5.0 and includes the qmode flags (big-endian, etc.) and CAF extended
+// channel layouts (including optional reordering). Even if there is no new
+// configuration, we still send the empty metadata block to signal a 5.0 file.
+
+static void write_new_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
+{
+    char *byteptr = wpmd->data = malloc (260);
+
+    wpmd->id = ID_NEW_CONFIG_BLOCK;
+
+    if (wpc->file_format || wpc->config.qmode || wpc->channel_layout) {
+        *byteptr++ = (char) wpc->file_format;
+        *byteptr++ = (char) wpc->config.qmode;
+
+        if (wpc->channel_layout) {
+            int nchans = wpc->channel_layout & 0xff;
+
+            *byteptr++ = (char) ((wpc->channel_layout & 0xff0000) >> 16);
+
+            if (wpc->channel_reordering || nchans != wpc->config.num_channels)
+                *byteptr++ = (char) nchans;
+
+            if (wpc->channel_reordering) {
+                int i, num_to_send = 0;
+
+                // to save space, don't send redundant reorder string bytes
+
+                for (i = 0; i < nchans; ++i)
+                    if (wpc->channel_reordering [i] != i)
+                        num_to_send = i + 1;
+
+                if (num_to_send) {
+                    memcpy (byteptr, wpc->channel_reordering, num_to_send);
+                    byteptr += num_to_send;
+                }
+            }
+        }
+    }
+
+    wpmd->byte_length = (int32_t)(byteptr - (char *) wpmd->data);
+}
+
+// Allocate room for and copy the non-standard sampling rate into the specified
+// metadata structure. We normally store the lower 3 bytes of the sampling rate,
+// unless 4 bytes are required (introduced in version 5). Note that this would
+// only be used when the sampling rate was not included in the table of 15
+// "standard" values.
 
 static void write_sample_rate (WavpackContext *wpc, WavpackMetadata *wpmd)
 {
@@ -371,6 +434,12 @@ static void write_sample_rate (WavpackContext *wpc, WavpackMetadata *wpmd)
     *byteptr++ = (char) (wpc->config.sample_rate);
     *byteptr++ = (char) (wpc->config.sample_rate >> 8);
     *byteptr++ = (char) (wpc->config.sample_rate >> 16);
+
+    // handle 4-byte sampling rates for scientific applications, etc.
+
+    if (wpc->config.sample_rate & 0x7f000000)
+        *byteptr++ = (char) (wpc->config.sample_rate >> 24) & 0x7f;
+
     wpmd->byte_length = (int32_t)(byteptr - (char *) wpmd->data);
 }
 
@@ -417,7 +486,10 @@ int pack_block (WavpackContext *wpc, int32_t *buffer)
         }
     }
 
-    if (!(flags & MONO_FLAG) && wpc->stream_version >= 0x410) {
+    // This code scans stereo data to check whether it can be stored as mono data
+    // (i.e., all L/R samples identical). Only available with MAX_STREAM_VERS.
+
+    if (!(flags & MONO_FLAG) && wpc->stream_version == MAX_STREAM_VERS) {
         int32_t lor = 0, diff = 0;
         int32_t *sptr, *dptr, i;
 
@@ -788,6 +860,45 @@ static void send_int32_data (WavpackStream *wps, int32_t *values, int32_t num_va
         }
 }
 
+void send_general_metadata (WavpackContext *wpc)
+{
+    WavpackStream *wps = wpc->streams [wpc->current_stream];
+    uint32_t flags = wps->wphdr.flags;
+    WavpackMetadata wpmd;
+
+    if ((flags & SRATE_MASK) == SRATE_MASK && wpc->config.sample_rate != 44100) {
+        write_sample_rate (wpc, &wpmd);
+        copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
+        free_metadata (&wpmd);
+    }
+
+    if ((flags & INITIAL_BLOCK) &&
+        (wpc->config.num_channels > 2 ||
+        wpc->config.channel_mask != 0x5 - wpc->config.num_channels)) {
+            write_channel_info (wpc, &wpmd);
+            copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
+            free_metadata (&wpmd);
+
+            if (wpc->channel_identities) {
+                write_channel_identities_info (wpc, &wpmd);
+                copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
+                free_metadata (&wpmd);
+            }
+    }
+
+    if ((flags & INITIAL_BLOCK) && !wps->sample_index) {
+        write_config_info (wpc, &wpmd);
+        copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
+        free_metadata (&wpmd);
+    }
+
+    if (flags & INITIAL_BLOCK) {
+        write_new_config_info (wpc, &wpmd);
+        copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
+        free_metadata (&wpmd);
+    }
+}
+
 // Pack an entire block of samples (either mono or stereo) into a completed
 // WavPack block. It is assumed that there is sufficient space for the
 // completed block at "wps->blockbuff" and that "wps->blockend" points to the
@@ -941,12 +1052,6 @@ static int pack_samples (WavpackContext *wpc, int32_t *buffer)
         copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
         free_metadata (&wpmd);
 
-        if ((flags & SRATE_MASK) == SRATE_MASK && wpc->config.sample_rate != 44100) {
-            write_sample_rate (wpc, &wpmd);
-            copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
-            free_metadata (&wpmd);
-        }
-
         if (flags & HYBRID_FLAG) {
             write_hybrid_profile (wps, &wpmd);
             copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
@@ -965,20 +1070,7 @@ static int pack_samples (WavpackContext *wpc, int32_t *buffer)
             free_metadata (&wpmd);
         }
 
-        if ((flags & INITIAL_BLOCK) &&
-            (wpc->config.num_channels > 2 ||
-            wpc->config.channel_mask != 0x5 - wpc->config.num_channels)) {
-                write_channel_info (wpc, &wpmd);
-                copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
-                free_metadata (&wpmd);
-        }
-
-        if ((flags & INITIAL_BLOCK) && !wps->sample_index) {
-            write_config_info (wpc, &wpmd);
-            copy_metadata (&wpmd, wps->blockbuff, wps->blockend);
-            free_metadata (&wpmd);
-        }
-
+        send_general_metadata (wpc);
         bs_open_write (&wps->wvbits, wps->blockbuff + ((WavpackHeader *) wps->blockbuff)->ckSize + 12, wps->blockend);
 
         if (wpc->wvc_flag) {
